@@ -31,6 +31,7 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <errno.h>
 
 #include <bluetooth/bluetooth.h>
@@ -44,7 +45,7 @@
 #include <dbus/dbus.h>
 #include <gdbus.h>
 
-#include "logging.h"
+#include "log.h"
 #include "textfile.h"
 
 #include "hcid.h"
@@ -80,7 +81,6 @@ struct bonding_req {
 	DBusConnection *conn;
 	DBusMessage *msg;
 	GIOChannel *io;
-	guint io_id;
 	guint listener_id;
 	struct btd_device *device;
 };
@@ -137,10 +137,14 @@ struct btd_device {
 
 	gboolean	trusted;
 	gboolean	paired;
+	gboolean	blocked;
 	gboolean	renewed_key;
 
 	gboolean	authorizing;
 	gint		ref;
+
+	gboolean	has_debug_key;
+	uint8_t		debug_key[16];
 };
 
 static uint16_t uuid_list[] = {
@@ -152,14 +156,6 @@ static uint16_t uuid_list[] = {
 
 static GSList *device_drivers = NULL;
 
-static DBusHandlerResult error_connection_attempt_failed(DBusConnection *conn,
-						DBusMessage *msg, int err)
-{
-	return error_common_reply(conn, msg,
-			ERROR_INTERFACE ".ConnectionAttemptFailed",
-			err > 0 ? strerror(err) : "Connection attempt failed");
-}
-
 static DBusHandlerResult error_failed(DBusConnection *conn,
 					DBusMessage *msg, const char * desc)
 {
@@ -182,7 +178,8 @@ static inline DBusMessage *no_such_adapter(DBusMessage *msg)
 
 static inline DBusMessage *in_progress(DBusMessage *msg, const char *str)
 {
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress", str);
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress",
+								"%s", str);
 }
 
 static void browse_request_free(struct browse_req *req)
@@ -246,7 +243,7 @@ static void device_free(gpointer user_data)
 	if (device->discov_timer)
 		g_source_remove(device->discov_timer);
 
-	debug("device_free(%p)", device);
+	DBG("%p", device);
 
 	g_free(device->authr);
 	g_free(device->path);
@@ -336,6 +333,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	boolean = device_is_trusted(device);
 	dict_append_entry(&dict, "Trusted", DBUS_TYPE_BOOLEAN, &boolean);
 
+	/* Blocked */
+	boolean = device->blocked;
+	dict_append_entry(&dict, "Blocked", DBUS_TYPE_BOOLEAN, &boolean);
+
 	/* Connected */
 	boolean = (device->handle != 0);
 	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN,
@@ -381,7 +382,7 @@ static DBusMessage *set_alias(DBusConnection *conn, DBusMessage *msg,
 	if (err < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(-err));
+				"%s", strerror(-err));
 
 	g_free(device->alias);
 	device->alias = g_str_equal(alias, "") ? NULL : g_strdup(alias);
@@ -413,7 +414,7 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 	if (err < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(-err));
+				"%s", strerror(-err));
 
 	device->trusted = value;
 
@@ -424,11 +425,155 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 	return dbus_message_new_method_return(msg);
 }
 
+static void driver_remove(struct btd_driver_data *driver_data,
+						struct btd_device *device)
+{
+	struct btd_device_driver *driver = driver_data->driver;
+
+	driver->remove(device);
+
+	device->drivers = g_slist_remove(device->drivers, driver_data);
+	g_free(driver_data);
+}
+
+static gboolean do_disconnect(gpointer user_data)
+{
+	struct btd_device *device = user_data;
+	disconnect_cp cp;
+	int dd;
+	uint16_t dev_id = adapter_get_dev_id(device->adapter);
+
+	device->disconn_timer = 0;
+
+	dd = hci_open_dev(dev_id);
+	if (dd < 0)
+		goto fail;
+
+	memset(&cp, 0, sizeof(cp));
+	cp.handle = htobs(device->handle);
+	cp.reason = HCI_OE_USER_ENDED_CONNECTION;
+
+	hci_send_cmd(dd, OGF_LINK_CTL, OCF_DISCONNECT,
+			DISCONNECT_CP_SIZE, &cp);
+
+	close(dd);
+
+fail:
+	return FALSE;
+}
+
+static int device_block(DBusConnection *conn, struct btd_device *device)
+{
+	int dev_id, dd, err;
+	bdaddr_t src;
+
+	if (device->blocked)
+		return 0;
+
+	dev_id = adapter_get_dev_id(device->adapter);
+
+	dd = hci_open_dev(dev_id);
+	if (dd < 0)
+		return -errno;
+
+	if (device->handle)
+		do_disconnect(device);
+
+	g_slist_foreach(device->drivers, (GFunc) driver_remove, device);
+
+	if (ioctl(dd, HCIBLOCKADDR, &device->bdaddr) < 0) {
+		err = -errno;
+		hci_close_dev(dd);
+		return err;
+	}
+
+	hci_close_dev(dd);
+
+	device->blocked = TRUE;
+
+	adapter_get_address(device->adapter, &src);
+
+	err = write_blocked(&src, &device->bdaddr, TRUE);
+	if (err < 0)
+		error("write_blocked(): %s (%d)", strerror(-err), -err);
+
+	device_set_temporary(device, FALSE);
+
+	emit_property_changed(conn, device->path, DEVICE_INTERFACE, "Blocked",
+					DBUS_TYPE_BOOLEAN, &device->blocked);
+
+	return 0;
+}
+
+static int device_unblock(DBusConnection *conn, struct btd_device *device,
+							gboolean silent)
+{
+	int dev_id, dd, err;
+	bdaddr_t src;
+
+	if (!device->blocked)
+		return 0;
+
+	dev_id = adapter_get_dev_id(device->adapter);
+
+	dd = hci_open_dev(dev_id);
+	if (dd < 0)
+		return -errno;
+
+	if (ioctl(dd, HCIUNBLOCKADDR, &device->bdaddr) < 0) {
+		err = -errno;
+		hci_close_dev(dd);
+		return err;
+	}
+
+	hci_close_dev(dd);
+
+	device->blocked = FALSE;
+
+	adapter_get_address(device->adapter, &src);
+
+	err = write_blocked(&src, &device->bdaddr, FALSE);
+	if (err < 0)
+		error("write_blocked(): %s (%d)", strerror(-err), -err);
+
+	if (!silent) {
+		emit_property_changed(conn, device->path,
+					DEVICE_INTERFACE, "Blocked",
+					DBUS_TYPE_BOOLEAN, &device->blocked);
+		device_probe_drivers(device, device->uuids);
+	}
+
+	return 0;
+}
+
+static DBusMessage *set_blocked(DBusConnection *conn, DBusMessage *msg,
+						gboolean value, void *data)
+{
+	struct btd_device *device = data;
+	int err;
+
+	if (value)
+		err = device_block(conn, device);
+	else
+		err = device_unblock(conn, device, FALSE);
+
+	switch (-err) {
+	case 0:
+		return dbus_message_new_method_return(msg);
+	case EINVAL:
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".NotSupported",
+					"Kernel lacks blacklist support");
+	default:
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"%s", strerror(-err));
+	}
+}
+
 static inline DBusMessage *invalid_args(DBusMessage *msg)
 {
-	return g_dbus_create_error(msg,
-			ERROR_INTERFACE ".InvalidArguments",
-			"Invalid arguments in method call");
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+					"Invalid arguments in method call");
 }
 
 static DBusMessage *set_property(DBusConnection *conn,
@@ -453,7 +598,6 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	if (g_str_equal("Trusted", property)) {
 		dbus_bool_t value;
-
 		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
 			return invalid_args(msg);
 		dbus_message_iter_get_basic(&sub, &value);
@@ -467,6 +611,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 		dbus_message_iter_get_basic(&sub, &alias);
 
 		return set_alias(conn, msg, alias, data);
+	} else if (g_str_equal("Blocked", property)) {
+		dbus_bool_t value;
+
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &value);
+
+		return set_blocked(conn, msg, value, data);
 	}
 
 	return invalid_args(msg);
@@ -476,7 +629,7 @@ static void discover_services_req_exit(DBusConnection *conn, void *user_data)
 {
 	struct browse_req *req = user_data;
 
-	debug("DiscoverServices requestor exited");
+	DBG("DiscoverServices requestor exited");
 
 	browse_request_cancel(req);
 }
@@ -631,42 +784,11 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 	return dbus_message_new_method_return(msg);
 }
 
-static gboolean do_disconnect(gpointer user_data)
-{
-	struct btd_device *device = user_data;
-	disconnect_cp cp;
-	int dd;
-	uint16_t dev_id = adapter_get_dev_id(device->adapter);
-
-	device->disconn_timer = 0;
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		goto fail;
-
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(device->handle);
-	cp.reason = HCI_OE_USER_ENDED_CONNECTION;
-
-	hci_send_cmd(dd, OGF_LINK_CTL, OCF_DISCONNECT,
-			DISCONNECT_CP_SIZE, &cp);
-
-	close(dd);
-
-fail:
-	return FALSE;
-}
-
 static void bonding_request_cancel(struct bonding_req *bonding)
 {
 	if (!bonding->io)
 		return;
 
-	if (bonding->io_id) {
-		g_source_remove(bonding->io_id);
-		bonding->io_id = 0;
-	}
-
 	g_io_channel_shutdown(bonding->io, TRUE, NULL);
 	g_io_channel_unref(bonding->io);
 	bonding->io = NULL;
@@ -880,7 +1002,7 @@ struct btd_device *device_create(DBusConnection *conn,
 	g_strdelimit(device->path, ":", '_');
 	g_free(address_up);
 
-	debug("Creating device %s", device->path);
+	DBG("Creating device %s", device->path);
 
 	if (g_dbus_register_interface(conn, device->path, DEVICE_INTERFACE,
 				device_methods, device_signals, NULL,
@@ -898,6 +1020,9 @@ struct btd_device *device_create(DBusConnection *conn,
 		device->alias = g_strdup(alias);
 	device->trusted = read_trust(&src, address, GLOBAL_TRUST);
 
+	if (read_blocked(&src, &device->bdaddr))
+		device_block(conn, device);
+
 	device->auth = 0xff;
 
 	if (read_link_key(&src, &device->bdaddr, NULL, NULL) == 0)
@@ -932,7 +1057,7 @@ void device_get_name(struct btd_device *device, char *name, size_t len)
 	strncpy(name, device->name, len);
 }
 
-static void device_remove_bonding(struct btd_device *device)
+void device_remove_bonding(struct btd_device *device)
 {
 	char filename[PATH_MAX + 1];
 	char srcaddr[18], dstaddr[18];
@@ -965,6 +1090,7 @@ static void device_remove_stored(struct btd_device *device)
 {
 	bdaddr_t src;
 	char addr[18];
+	DBusConnection *conn = get_dbus_connection();
 
 	adapter_get_address(device->adapter, &src);
 	ba2str(&device->bdaddr, addr);
@@ -974,14 +1100,18 @@ static void device_remove_stored(struct btd_device *device)
 	delete_entry(&src, "profiles", addr);
 	delete_entry(&src, "trusts", addr);
 	delete_all_records(&src, &device->bdaddr);
+
+	if (device->blocked)
+		device_unblock(conn, device, TRUE);
 }
 
 void device_remove(struct btd_device *device, gboolean remove_stored)
 {
-	GSList *list;
-	struct btd_device_driver *driver;
 
-	debug("Removing device %s", device->path);
+	DBG("Removing device %s", device->path);
+
+	if (device->agent)
+		agent_free(device->agent);
 
 	if (device->bonding)
 		device_cancel_bonding(device, HCI_OE_USER_ENDED_CONNECTION);
@@ -995,13 +1125,9 @@ void device_remove(struct btd_device *device, gboolean remove_stored)
 	if (remove_stored)
 		device_remove_stored(device);
 
-	for (list = device->drivers; list; list = list->next) {
-		struct btd_driver_data *driver_data = list->data;
-		driver = driver_data->driver;
-
-		driver->remove(device);
-		g_free(driver_data);
-	}
+	g_slist_foreach(device->drivers, (GFunc) driver_remove, device);
+	g_slist_free(device->drivers);
+	device->drivers = NULL;
 
 	btd_device_unref(device);
 }
@@ -1096,7 +1222,12 @@ void device_probe_drivers(struct btd_device *device, GSList *profiles)
 	GSList *list;
 	int err;
 
-	debug("Probe drivers for %s", device->path);
+	if (device->blocked) {
+		DBG("Skipping drivers for blocked device %s", device->path);
+		goto add_uuids;
+	}
+
+	DBG("Probe drivers for %s", device->path);
 
 	for (list = device_drivers; list; list = list->next) {
 		struct btd_device_driver *driver = list->data;
@@ -1125,6 +1256,7 @@ void device_probe_drivers(struct btd_device *device, GSList *profiles)
 		g_slist_free(probe_uuids);
 	}
 
+add_uuids:
 	for (list = profiles; list; list = list->next) {
 		GSList *l = g_slist_find_custom(device->uuids, list->data,
 						(GCompareFunc) strcasecmp);
@@ -1135,12 +1267,6 @@ void device_probe_drivers(struct btd_device *device, GSList *profiles)
 						g_strdup(list->data),
 						(GCompareFunc) strcasecmp);
 	}
-
-	if (device->tmp_records) {
-		sdp_list_free(device->tmp_records,
-				(sdp_free_func_t) sdp_record_free);
-		device->tmp_records = NULL;
-	}
 }
 
 static void device_remove_drivers(struct btd_device *device, GSList *uuids)
@@ -1157,7 +1283,7 @@ static void device_remove_drivers(struct btd_device *device, GSList *uuids)
 
 	records = read_records(&src, &device->bdaddr);
 
-	debug("Remove drivers for %s", device->path);
+	DBG("Remove drivers for %s", device->path);
 
 	for (list = device->drivers; list; list = next) {
 		struct btd_driver_data *driver_data = list->data;
@@ -1171,7 +1297,7 @@ static void device_remove_drivers(struct btd_device *device, GSList *uuids)
 						(GCompareFunc) strcasecmp))
 				continue;
 
-			debug("UUID %s was removed from device %s",
+			DBG("UUID %s was removed from device %s",
 							*uuid, dstaddr);
 
 			driver->remove(device);
@@ -1215,7 +1341,7 @@ static void services_changed(struct btd_device *device)
 		uuids[i] = l->data;
 
 	emit_array_property_changed(conn, device->path, DEVICE_INTERFACE,
-					"UUIDs", DBUS_TYPE_STRING, &uuids);
+					"UUIDs", DBUS_TYPE_STRING, &uuids, i);
 
 	g_free(uuids);
 }
@@ -1254,7 +1380,7 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 
 		/* Check for empty service classes list */
 		if (svcclass == NULL) {
-			debug("Skipping record with no service classes");
+			DBG("Skipping record with no service classes");
 			continue;
 		}
 
@@ -1361,15 +1487,15 @@ static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 
 	update_services(req, recs);
 
-	if (device->tmp_records && req->records) {
+	if (device->tmp_records)
 		sdp_list_free(device->tmp_records,
 					(sdp_free_func_t) sdp_record_free);
-		device->tmp_records = req->records;
-		req->records = NULL;
-	}
+
+	device->tmp_records = req->records;
+	req->records = NULL;
 
 	if (!req->profiles_added && !req->profiles_removed) {
-		debug("%s: No service update", device->path);
+		DBG("%s: No service update", device->path);
 		goto send_reply;
 	}
 
@@ -1390,7 +1516,7 @@ send_reply:
 
 	if (dbus_message_is_method_call(req->msg, DEVICE_INTERFACE,
 					"DiscoverServices"))
-		discover_services_reply(req, err, req->records);
+		discover_services_reply(req, err, device->tmp_records);
 	else if (dbus_message_is_method_call(req->msg, ADAPTER_INTERFACE,
 						"CreatePairedDevice"))
 		create_device_reply(device, req);
@@ -1544,15 +1670,10 @@ struct agent *device_get_agent(struct btd_device *device)
 	if (!device)
 		return NULL;
 
-	return  device->agent;
-}
+	if (device->agent)
+		return device->agent;
 
-void device_set_agent(struct btd_device *device, struct agent *agent)
-{
-	if (!device)
-		return;
-
-	device->agent = agent;
+	return adapter_get_agent(device->adapter);
 }
 
 gboolean device_is_busy(struct btd_device *device)
@@ -1610,7 +1731,7 @@ static gboolean start_discovery(gpointer user_data)
 	return FALSE;
 }
 
-DBusMessage *new_authentication_return(DBusMessage *msg, uint8_t status)
+static DBusMessage *new_authentication_return(DBusMessage *msg, int status)
 {
 	switch (status) {
 	case 0x00: /* success */
@@ -1681,9 +1802,6 @@ static void bonding_request_free(struct bonding_req *bonding)
 	if (bonding->conn)
 		dbus_connection_unref(bonding->conn);
 
-	if (bonding->io_id)
-		g_source_remove(bonding->io_id);
-
 	if (bonding->io)
 		g_io_channel_unref(bonding->io);
 
@@ -1720,7 +1838,7 @@ static void device_agent_removed(struct agent *agent, void *user_data)
 {
 	struct btd_device *device = user_data;
 
-	device_set_agent(device, NULL);
+	device->agent = NULL;
 
 	if (device->authr)
 		device->authr->agent = NULL;
@@ -1736,7 +1854,7 @@ static struct bonding_req *bonding_request_new(DBusConnection *conn,
 	const char *name = dbus_message_get_sender(msg);
 	struct agent *agent;
 
-	debug("%s: requesting bonding", device->path);
+	DBG("%s: requesting bonding", device->path);
 
 	if (!agent_path)
 		goto proceed;
@@ -1752,7 +1870,7 @@ static struct bonding_req *bonding_request_new(DBusConnection *conn,
 
 	device->agent = agent;
 
-	debug("Temporary agent registered for %s at %s:%s",
+	DBG("Temporary agent registered for %s at %s:%s",
 			device->path, name, agent_path);
 
 proceed:
@@ -1764,63 +1882,19 @@ proceed:
 	return bonding;
 }
 
-static gboolean bonding_io_cb(GIOChannel *io, GIOCondition cond,
-							gpointer user_data)
+static int device_authentication_requested(struct btd_device *device,
+						int handle)
 {
-	struct btd_device *device = user_data;
-	DBusMessage *reply;
-
-	if (!device->bonding)
-		return FALSE;
-
-	reply = new_authentication_return(device->bonding->msg,
-					HCI_CONNECTION_TERMINATED);
-	g_dbus_send_message(device->bonding->conn, reply);
-
-	bonding_request_free(device->bonding);
-
-	return FALSE;
-}
-
-static void bonding_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
-{
-	struct btd_device *device = user_data;
 	struct hci_request rq;
 	auth_requested_cp cp;
 	evt_cmd_status rp;
 	int dd;
-	uint16_t handle;
-
-	if (!device->bonding) {
-		if (!err)
-			g_io_channel_shutdown(io, TRUE, NULL);
-		return;
-	}
-
-	if (err) {
-		error("%s", err->message);
-		error_connection_attempt_failed(device->bonding->conn,
-						device->bonding->msg,
-						ENETDOWN);
-		goto cleanup;
-	}
-
-	if (!bt_io_get(io, BT_IO_L2RAW, &err,
-			BT_IO_OPT_HANDLE, &handle,
-			BT_IO_OPT_INVALID)) {
-		error("Unable to get connection handle: %s", err->message);
-		error_connection_attempt_failed(device->bonding->conn,
-						device->bonding->msg,
-						ENETDOWN);
-		g_error_free(err);
-		goto failed;
-	}
 
 	dd = hci_open_dev(adapter_get_dev_id(device->adapter));
 	if (dd < 0) {
-		DBusMessage *reply = no_such_adapter(device->bonding->msg);
-		g_dbus_send_message(device->bonding->conn, reply);
-		goto failed;
+		int err = -errno;
+		error("Unable to open adapter: %s(%d)", strerror(-err), -err);
+		return err;
 	}
 
 	memset(&rp, 0, sizeof(rp));
@@ -1838,44 +1912,67 @@ static void bonding_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
 	rq.event  = EVT_CMD_STATUS;
 
 	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
+		int err = -errno;
 		error("Unable to send HCI request: %s (%d)",
-					strerror(errno), errno);
-		error_failed_errno(device->bonding->conn, device->bonding->msg,
-				errno);
+					strerror(-err), -err);
 		hci_close_dev(dd);
-		goto failed;
+		return err;
 	}
 
 	if (rp.status) {
 		error("HCI_Authentication_Requested failed with status 0x%02x",
 				rp.status);
-		error_failed_errno(device->bonding->conn, device->bonding->msg,
-				bt_error(rp.status));
 		hci_close_dev(dd);
+		return rp.status;
+	}
+
+	info("Authentication requested");
+
+	hci_close_dev(dd);
+	return 0;
+}
+
+static void bonding_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
+{
+	struct btd_device *device = user_data;
+	uint16_t handle;
+	int status;
+
+	if (!device->bonding) {
+		if (!err)
+			g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (err)
+		/* Wait proper error to be propagated by bonding complete */
+		return;
+
+	if (!bt_io_get(io, BT_IO_L2RAW, &err,
+			BT_IO_OPT_HANDLE, &handle,
+			BT_IO_OPT_INVALID)) {
+		error("Unable to get connection handle: %s", err->message);
+		g_error_free(err);
+		status = -errno;
 		goto failed;
 	}
 
-	hci_close_dev(dd);
-
-	device->bonding->io_id = g_io_add_watch(io,
-					G_IO_NVAL | G_IO_HUP | G_IO_ERR,
-					bonding_io_cb, device);
+	status = device_authentication_requested(device, handle);
+	if (status != 0)
+		goto failed;
 
 	return;
 
 failed:
 	g_io_channel_shutdown(io, TRUE, NULL);
-
-cleanup:
-	device->bonding->io_id = 0;
-	bonding_request_free(device->bonding);
+	device_cancel_bonding(device, status);
 }
 
 static void create_bond_req_exit(DBusConnection *conn, void *user_data)
 {
 	struct btd_device *device = user_data;
 
-	debug("%s: requestor exited before bonding was completed", device->path);
+	DBG("%s: requestor exited before bonding was completed", device->path);
 
 	if (device->authr)
 		device_cancel_authentication(device, FALSE);
@@ -1899,6 +1996,7 @@ DBusMessage *device_create_bonding(struct btd_device *device,
 	bdaddr_t src;
 	GError *err = NULL;
 	GIOChannel *io;
+	BtIOSecLevel sec_level;
 
 	adapter_get_address(adapter, &src);
 	ba2str(&src, srcaddr);
@@ -1919,18 +2017,25 @@ DBusMessage *device_create_bonding(struct btd_device *device,
 				"Bonding already exists");
 	}
 
+	/* If our IO capability is NoInputNoOutput use medium security
+	 * level (i.e. don't require MITM protection) else use high
+	 * security level */
+	if (capability == 0x03)
+		sec_level = BT_IO_SEC_MEDIUM;
+	else
+		sec_level = BT_IO_SEC_HIGH;
 
 	io = bt_io_connect(BT_IO_L2RAW, bonding_connect_cb, device,
 				NULL, &err,
 				BT_IO_OPT_SOURCE_BDADDR, &src,
 				BT_IO_OPT_DEST_BDADDR, &device->bdaddr,
-				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_HIGH,
+				BT_IO_OPT_SEC_LEVEL, sec_level,
 				BT_IO_OPT_INVALID);
 	if (io == NULL) {
 		DBusMessage *reply;
 		reply = g_dbus_create_error(msg,
 				ERROR_INTERFACE ".ConnectionAttemptFailed",
-				err->message);
+				"%s", err->message);
 		error("bt_io_connect: %s", err->message);
 		g_error_free(err);
 		return reply;
@@ -1972,8 +2077,11 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 	if (auth && auth->type == AUTH_TYPE_NOTIFY && auth->agent)
 		agent_cancel(auth->agent);
 
-	if (status)
-		goto failed;
+	if (status) {
+		device_cancel_authentication(device, TRUE);
+		device_cancel_bonding(device, status);
+		return;
+	}
 
 	device->auth = 0xff;
 
@@ -2007,7 +2115,7 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 			/* If we are not initiators and there is no currently
 			 * active discovery or discovery timer, set discovery
 			 * timer */
-			debug("setting timer for reverse service discovery");
+			DBG("setting timer for reverse service discovery");
 			device->discov_timer = g_timeout_add_seconds(
 							DISCOVERY_TIMER,
 							start_discovery,
@@ -2016,11 +2124,6 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 	}
 
 	device_set_paired(device, TRUE);
-
-	return;
-
-failed:
-	device_cancel_bonding(device, status);
 }
 
 gboolean device_is_creating(struct btd_device *device, const char *sender)
@@ -2067,7 +2170,7 @@ void device_cancel_bonding(struct btd_device *device, uint8_t status)
 	if (!bonding)
 		return;
 
-	debug("%s: canceling bonding request", device->path);
+	DBG("%s: canceling bonding request", device->path);
 
 	if (device->authr)
 		device_cancel_authentication(device, FALSE);
@@ -2079,8 +2182,8 @@ void device_cancel_bonding(struct btd_device *device, uint8_t status)
 	bonding_request_free(bonding);
 }
 
-static void pincode_cb(struct agent *agent, DBusError *err, const char *pincode,
-			void *data)
+static void pincode_cb(struct agent *agent, DBusError *err,
+					const char *pincode, void *data)
 {
 	struct authentication_req *auth = data;
 	struct btd_device *device = auth->device;
@@ -2110,8 +2213,8 @@ static void confirm_cb(struct agent *agent, DBusError *err, void *data)
 	device->authr->agent = NULL;
 }
 
-static void passkey_cb(struct agent *agent, DBusError *err, uint32_t passkey,
-			void *data)
+static void passkey_cb(struct agent *agent, DBusError *err,
+						uint32_t passkey, void *data)
 {
 	struct authentication_req *auth = data;
 	struct btd_device *device = auth->device;
@@ -2127,19 +2230,15 @@ static void passkey_cb(struct agent *agent, DBusError *err, uint32_t passkey,
 }
 
 int device_request_authentication(struct btd_device *device, auth_type_t type,
-				uint32_t passkey, void *cb)
+						uint32_t passkey, void *cb)
 {
 	struct authentication_req *auth;
 	struct agent *agent;
-	int ret;
+	int err;
 
-	debug("%s: requesting agent authentication", device->path);
-
-	agent = device->agent;
-
-	if (!agent)
-		agent = adapter_get_agent(device->adapter);
+	DBG("%s: requesting agent authentication", device->path);
 
+	agent = device_get_agent(device);
 	if (!agent) {
 		error("No agent available for %u request", type);
 		return -EPERM;
@@ -2154,34 +2253,34 @@ int device_request_authentication(struct btd_device *device, auth_type_t type,
 
 	switch (type) {
 	case AUTH_TYPE_PINCODE:
-		ret = agent_request_pincode(agent, device, pincode_cb,
+		err = agent_request_pincode(agent, device, pincode_cb,
 								auth, NULL);
 		break;
 	case AUTH_TYPE_PASSKEY:
-		ret = agent_request_passkey(agent, device, passkey_cb,
+		err = agent_request_passkey(agent, device, passkey_cb,
 								auth, NULL);
 		break;
 	case AUTH_TYPE_CONFIRM:
-		ret = agent_request_confirmation(agent, device, passkey,
+		err = agent_request_confirmation(agent, device, passkey,
 						confirm_cb, auth, NULL);
 		break;
 	case AUTH_TYPE_NOTIFY:
-		ret = agent_display_passkey(agent, device, passkey);
+		err = agent_display_passkey(agent, device, passkey);
 		break;
 	case AUTH_TYPE_AUTO:
-		ret = 0;
+		err = 0;
 		break;
 	default:
-		ret = -EINVAL;
+		err = -EINVAL;
 	}
 
-	if (ret < 0) {
+	if (err < 0) {
 		error("Failed requesting authentication");
 		g_free(auth);
 		device->authr = NULL;
 	}
 
-	return ret;
+	return err;
 }
 
 static void cancel_authentication(struct authentication_req *auth)
@@ -2226,7 +2325,7 @@ void device_cancel_authentication(struct btd_device *device, gboolean aborted)
 	if (!auth)
 		return;
 
-	debug("%s: canceling authentication request", device->path);
+	DBG("%s: canceling authentication request", device->path);
 
 	if (auth->agent)
 		agent_cancel(auth->agent);
@@ -2280,12 +2379,17 @@ void btd_device_add_uuid(struct btd_device *device, const char *uuid)
 }
 
 const sdp_record_t *btd_device_get_record(struct btd_device *device,
-						const char *uuid)
+							const char *uuid)
 {
 	bdaddr_t src;
 
-	if (device->tmp_records)
-		return find_record_in_list(device->tmp_records, uuid);
+	if (device->tmp_records) {
+		const sdp_record_t *record;
+
+		record = find_record_in_list(device->tmp_records, uuid);
+		if (record != NULL)
+			return record;
+	}
 
 	adapter_get_address(device->adapter, &src);
 
@@ -2296,6 +2400,30 @@ const sdp_record_t *btd_device_get_record(struct btd_device *device,
 	return find_record_in_list(device->tmp_records, uuid);
 }
 
+gboolean device_set_debug_key(struct btd_device *device, uint8_t *key)
+{
+	if (key == NULL) {
+		device->has_debug_key = FALSE;
+		return TRUE;
+	}
+
+	memcpy(device->debug_key, key, 16);
+	device->has_debug_key = TRUE;
+
+	return TRUE;
+}
+
+gboolean device_get_debug_key(struct btd_device *device, uint8_t *key)
+{
+	if (!device->has_debug_key)
+		return FALSE;
+
+	if (key != NULL)
+		memcpy(key, device->debug_key, 16);
+
+	return TRUE;
+}
+
 int btd_register_device_driver(struct btd_device_driver *driver)
 {
 	device_drivers = g_slist_append(device_drivers, driver);
@@ -2312,7 +2440,7 @@ struct btd_device *btd_device_ref(struct btd_device *device)
 {
 	device->ref++;
 
-	debug("btd_device_ref(%p): ref=%d", device, device->ref);
+	DBG("%p: ref=%d", device, device->ref);
 
 	return device;
 }
@@ -2324,7 +2452,7 @@ void btd_device_unref(struct btd_device *device)
 
 	device->ref--;
 
-	debug("btd_device_unref(%p): ref=%d", device, device->ref);
+	DBG("%p: ref=%d", device, device->ref);
 
 	if (device->ref > 0)
 		return;
